/* tapplay - TAP file playback for sd2iec hardware
   Copyright (C) 2009  Ingo Korb <ingo@akana.de>

   Inspiration and low-level SD/MMC access based on code from MMC2IEC
     by Lars Pontoppidan et al., see sdcard.c|h and config.h.

   FAT filesystem access based on code from ChaN and Jim Brain, see ff.c|h.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License only.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   cli.c: The command line interface

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <util/delay.h>
#include "config.h"
#include "avrcompat.h"
#include "diskio.h"
#include "ff.h"
#include "tapsend.h"
#include "timer.h"
#include "uart.h"

#define MAX_LINE 250

/* Variables */
static char cmdbuffer[MAX_LINE+1];
static char *curchar;

FATFS fs;

/* Word lists */
static PROGMEM char command_words[] =
  "cd\0reset\0send\0dir\0ls\0queue\0";
enum { CMD_CD = 0, CMD_RESET, CMD_SEND, CMD_DIR, CMD_LS, CMD_QUEUE };

static PROGMEM char queue_words[] =
  "add\0delete\0show\0clear\0send\0";
enum { CMD_Q_ADD = 0, CMD_Q_DELETE, CMD_Q_SHOW, CMD_Q_CLEAR, CMD_Q_SEND };

typedef struct queue_s {
  FIL fd;
  char *name;
  struct queue_s *next;
} queue_t;

static queue_t *queuehead;

/* ------------------------------------------------------------------------- */
/*   Parse functions                                                         */
/* ------------------------------------------------------------------------- */

/* Skip spaces at curchar */
static uint8_t skip_spaces(void) {
  uint8_t res = (*curchar == ' ' || *curchar == 0);

  while (*curchar == ' ')
    curchar++;

  return res;
}

/* Parse the string in curchar for an integer with bounds [lower,upper] */
static int32_t parse_unsigned(uint32_t lower, uint32_t upper) {
  char *end;
  uint32_t result;

  if (strlen(curchar) == 1 && *curchar == '?') {
    dprintf("Number between %ld and %ld expected\n",lower,upper);
    return -2;
  }

  result = strtoul(curchar, &end, 10);
  if ((*end != ' ' && *end != 0) || errno != 0) {
    dprintf("Invalid numeric argument\n");
    return -1;
  }

  curchar = end;
  skip_spaces();

  if (result < lower || result > upper) {
    dprintf("Numeric argument out of range (%ld..%ld)\n",lower,upper);
    return -1;
  }

  return result;
}

/* Parse the string starting with curchar for a word in wordlist */
static int8_t parse_wordlist(char *wordlist) {
  uint8_t i, matched;
  char *cur, *ptr;
  char c;

  i = 0;
  ptr = wordlist;

  // Command list on "?"
  if (strlen(curchar) == 1 && *curchar == '?') {
    dprintf("Commands available: \n ");
    while (1) {
      c = pgm_read_byte(ptr++);
      if (c == 0) {
        if (pgm_read_byte(ptr) == 0) {
          uart_putcrlf();
          return -2;
        } else {
          uart_putcrlf();
          uart_putc(' ');
        }
      } else
        uart_putc(c);
    }
  }

  while (1) {
    cur = curchar;
    matched = 1;
    c = pgm_read_byte(ptr);
    do {
      // If current word list character is \0: No match found
      if (c == 0) {
        dprintf("Unknown word: %s\n",curchar);
        return -1;
      }

      if (tolower(c) != tolower(*cur)) {
        // Check for end-of-word
        if (cur != curchar && (*cur == ' ' || *cur == 0)) {
          // Partial match found, return that
          break;
        } else {
          matched = 0;
          break;
        }
      }
      ptr++;
      cur++;
      c = pgm_read_byte(ptr);
    } while (c != 0);
    
    if (matched) {
      char *tmp = curchar;

      curchar = cur;
      // Return match only if whitespace or end-of-string follows
      // (avoids mismatching partial words)
      if (skip_spaces()) {
        return i;
      } else {
        dprintf("Unknown word: %s\n(use ? for help)\n");
        return -1;
      }
    } else {
      // Try next word in list
      i++;
      while (pgm_read_byte(ptr++) != 0) ;
    }
  }
}

/* Read a line from serial, uses cmdbuffer as storage */
static char *getline(char *prompt) {
  int i=0;
  char c;

  uart_putcrlf();
  uart_puts_P(prompt);
  memset(cmdbuffer,0,sizeof(cmdbuffer));

  while (1) {
    c = uart_getc();
    if (c == 13)
      break;

    if (c == 27 || c == 3) {
      uart_putc('\\');
      uart_putcrlf();
      uart_puts_P(prompt);
      i = 0;
      memset(cmdbuffer,0,sizeof(cmdbuffer));
      continue;
    }

    if (c == 127 || c == 8) {
      if (i > 0) {
        i--;
        uart_putc(8);   // backspace
        uart_putc(' '); // erase character
        uart_putc(8);   // backspace
      } else
        continue;
    } else {
      if (i < sizeof(cmdbuffer)-1) {
        cmdbuffer[i++] = c;
        uart_putc(c);
      }
    }
  }
  cmdbuffer[i] = 0;
  return cmdbuffer;
}

/* ------------------------------------------------------------------------- */
/*   Queue function                                                          */
/* ------------------------------------------------------------------------- */

static void cmd_q_add(void) {
  queue_t *newq,*cur;
  FRESULT res;

  /* Allocate new entry */
  newq = malloc(sizeof(queue_t));
  if (newq == NULL) {
    dprintf("ERROR: Not enough memory to add a queue entry!\n");
    return;
  }

  memset(newq,0,sizeof(queue_t));
  res = f_open(&fs, &newq->fd, (const UCHAR *)curchar, FA_READ);
  if (res != FR_OK) {
    dprintf("f_open \"%s\" failed, result %d\n",curchar,res);
    free(newq);
    return;
  }

  newq->name = malloc(strlen(curchar)+1);
  if (newq->name == NULL) {
    dprintf("Warning: Not enough memory to remember the file name.\n");
  } else {
    strcpy(newq->name, curchar);
  }

  /* Add entry to queue chain */
  if (queuehead == NULL) {
    queuehead = newq;
  } else {
    cur = queuehead;
    while (cur->next != NULL)
      cur = cur->next;
    cur->next = newq;
  }
  dprintf("Added \"%s\" to the queue\n",curchar);
}

static void cmd_q_delete(void) {
  queue_t *cur = queuehead;
  queue_t *prev;
  unsigned int i= 0;
  int32_t target;

  if (queuehead == NULL) {
    dprintf("Queue is already empty!\n");
    return;
  }

  /* Count number of entries in the queue */
  while (cur) {
    i++;
    cur = cur->next;
  }

  target = parse_unsigned(1,i);
  if (target < 0)
    return;

  if (target == 1) {
    cur = queuehead;
    queuehead = queuehead->next;
  } else {
    cur = queuehead;
    i = 1;
    prev = NULL;
    while (i < target) {
      prev = cur;
      cur = cur->next;
      i++;
    }

    prev->next = cur->next;
  }
  free(cur->name);
  free(cur);
}

static void cmd_q_show(void) {
  queue_t *cur = queuehead;
  unsigned int i = 0;

  dprintf("Queue contents:\n");
  while (cur) {
    if (cur->name)
      dprintf("%3d: %s\n",++i,cur->name);
    else
      dprintf("%3d: (unknown)\n",++i);
    cur = cur->next;
  }
}

static void cmd_q_clear(void) {
  queue_t *cur = queuehead;
  queue_t *next;

  while (cur) {
    free(cur->name);
    next = cur->next;
    free(cur);
    cur = next;
  }
  queuehead = NULL;
  dprintf("Queue cleared.\n");
}

static void cmd_q_send(void) {
  queue_t *cur = queuehead;

  while (cur) {
    if (cur->name)
      dprintf("Sending \"%s\"\n",cur->name);
    else
      dprintf("Sending unknown file\n");

    send_file(&cur->fd);

    cur = cur->next;

    /* Add a small delay between files */
    if (cur)
      for (unsigned char i=0;i<200;i++)
        _delay_ms(10);

  }
}

/* ------------------------------------------------------------------------- */
/*   Command functions                                                       */
/* ------------------------------------------------------------------------- */

static void (*restart_call)(void) = 0;

/* Reset AVR */
static void cmd_reset(void) {
  cli();
  restart_call();
}

/* Show the contents of the current directory */
static void cmd_show_directory(void) {
  FRESULT res;
  DIR dh;
  FILINFO finfo;
  uint8_t *name;

  res = l_opendir(&fs, fs.curr_dir, &dh);
  if (res != FR_OK) {
    dprintf("l_opendir failed, result %d\n",res);
    return;
  }

  finfo.lfn = databuffer;

  do {
    /* Read the next entry */
    res = f_readdir(&dh, &finfo);
    if (res != FR_OK) {
      dprintf("f_readdir failed, result %d\n",res);
      return;
    }

    /* Abort if none was found */
    if (!finfo.fname[0])
      break;

    /* Skip volume labels */
    if (finfo.fattrib & AM_VOL)
      continue;

    /* Select between LFN and 8.3 name */
    if (finfo.lfn[0])
      name = finfo.lfn;
    else {
      name = finfo.fname;
      strlwr((char *)name);
    }

#if 0
    if (finfo.fattrib & AM_DIR) {
      dprintf("DIR ");
    } else {
      dprintf("    ");
    }
#endif

    dprintf("%s",name);

    /* Directory indicator (Unix-style) */
    if (finfo.fattrib & AM_DIR)
      uart_putc('/');

    uart_putcrlf();
  } while (finfo.fname[0]);
}

/* Send a single file */
static void cmd_send_file(void) {
  FRESULT res;
  FIL fd;

  res = f_open(&fs, &fd, (const UCHAR *)curchar, FA_READ);
  if (res != FR_OK) {
    dprintf("f_open %s failed, result %d\n",curchar,res);
  } else {
    dprintf("Sending %s, press escape to abort\n",curchar);
    send_file(&fd);
  }
}

#if 0
/* Send multiple files */
static void cmd_send_multiple(void) {
  char *next = curchar;
  char *last = curchar+strlen(curchar);

  while (next != last) {
    skip_spaces();
    if (strlen(curchar) == 0)
      return;

    next = curchar;

    /* Find name seperator */
    while (*next != '!' && *next != 0) next++;
    
    /* Null it, strip spaces before it */
    *next-- = 0;
    while (*next == ' ')
      *next-- = 0;

    /* send current file */
    cmd_send_file();

    /* search for next name */
    next++;
    while (*next == 0 && next != last)
      next++;

    curchar = next;
    for (char i=0;i<20;i++)
      _delay_ms(100);
  }
}
#endif

static void cmd_queue(void)  {
  int8_t cmd = parse_wordlist(queue_words);

  if (cmd < 0)
    return;

  switch (cmd) {
  case CMD_Q_ADD:
    cmd_q_add();
    break;

  case CMD_Q_DELETE:
    cmd_q_delete();
    break;

  case CMD_Q_SHOW:
    cmd_q_show();
    break;

  case CMD_Q_CLEAR:
    cmd_q_clear();
    break;

  case CMD_Q_SEND:
    cmd_q_send();
    break;
  }
}

/* ------------------------------------------------------------------------- */
/*   CLI interface functions                                                 */
/* ------------------------------------------------------------------------- */

void cli_init(void) {
  FRESULT res;

  dprintf("\n\ntapsend " VERSION " started\n");

  res = f_mount(0,&fs);
  if (f_mount(0,&fs) != FR_OK) {
    dprintf("WARNING: Failed to mount card, result %d\n",res);
  }
}

void cli_loop(void) {
  while (1) {
    curchar = getline(PSTR(">"));
    uart_putcrlf();

    /* Process medium changes before executing the command */
    if (disk_state != DISK_OK && disk_state != DISK_REMOVED) {
      FRESULT res;

      dprintf("Medium changed... ");
      res = f_mount(0,&fs);
      if (res != FR_OK) {
        dprintf("Failed to mount new medium, result %d\n",res);
      } else {
        dprintf("Ok\n");
      }

      if (queuehead)
        cmd_q_clear();
    }

    /* Remove whitespace */
    while (*curchar == ' ') curchar++;
    while (strlen(curchar) > 0 && curchar[strlen(curchar)-1] == ' ')
      curchar[strlen(curchar)-1] = 0;

    /* Ignore empty lines */
    if (strlen(curchar) == 0)
      continue;

    /* Parse command */
    int8_t command = parse_wordlist(command_words);
    if (command < 0)
      continue;

    FRESULT res;

    switch (command) {
    case CMD_CD:
      if (strlen(curchar) == 0) {
        dprintf("You are here and I don't know where that is.\n");
        break;
      }

      res = f_chdir(&fs, (const UCHAR *)curchar);
      if (res != FR_OK) {
        dprintf("chdir %s failed with result %d\n",curchar,res);
      } else {
        uart_puts_P(PSTR("Ok.\n\r"));
      }
      break;

    case CMD_RESET:
      cmd_reset();
      break;

    case CMD_SEND:
      cmd_send_file();
      break;

    case CMD_DIR:
    case CMD_LS:
      cmd_show_directory();
      break;

    case CMD_QUEUE:
      cmd_queue();
      break;
    }

  }
}
